#include <unistd.h>
#include <sys/timerfd.h>
#include <stdint.h>
#include <stdio.h>

#include <boost/bind.hpp>
#include <boost/noncopyable.hpp>

class PeriodicTimer : boost::noncopyable
{
private:
    class ITimerCallback
    {
    public:
        virtual ~ITimerCallback() {}
        virtual void OnTimer() = 0;
    };

    template <class Callback>
    class TimerCallbackFunctor : public ITimerCallback
    {
    public:
        TimerCallbackFunctor(const Callback& callback) :
          m_callback(callback)
        {
        }

        void OnTimer()
        {
            m_callback();
        }

    private:
        const Callback m_callback;
    };

public:
    PeriodicTimer();

    template <class Callable>
    void start(int interval, const Callable& timerFunction)
    {
        TimerCallbackFunctor<Callable> callback(timerFunction);
        startInternal(interval, callback);
    }

    void stop();

private:
    void startInternal(int interval, ITimerCallback& callback);

    class PeriodicTimerImpl;
    std::auto_ptr<PeriodicTimerImpl> m_pPeriodicTimer;
};

class PeriodicTimer::PeriodicTimerImpl
{
public:
    PeriodicTimerImpl() : m_timerfd(-1)
    {
    }

    void start(int interval, ITimerCallback& callback)
    {
        struct timespec timespec;
        timespec.tv_sec = interval / 1000;
        timespec.tv_nsec = (interval % 1000) * 1000000;

        struct itimerspec timerspec, oldtimespec;
        timerspec.it_value = timespec;
        timerspec.it_interval = timespec;

        m_timerfd = ::timerfd_create(CLOCK_MONOTONIC, 0);
        ::timerfd_settime(m_timerfd, 0, &timerspec, &oldtimespec);
  
        int64_t numberOfIntervals = 0;
        while (::read(m_timerfd, &numberOfIntervals, sizeof(int64_t)) == sizeof(int64_t))
        {
            callback.OnTimer();
        }

        m_timerfd = -1;
    }

    void stop()
    {
        ::close(m_timerfd);
    }

private:
    int m_timerfd;
};

PeriodicTimer::PeriodicTimer() :
    m_pPeriodicTimer(new PeriodicTimer::PeriodicTimerImpl())
{
}

void PeriodicTimer::startInternal(int interval, ITimerCallback& callback)
{
    return m_pPeriodicTimer->start(interval, callback);
}

void PeriodicTimer::stop()
{
    return m_pPeriodicTimer->stop();
}

class PeriodicTimerTests
{
public:
    void Tick()
    {
        m_timer.stop();
    }

    void shouldBeAbleToRestartTimer();

    PeriodicTimer m_timer;
};

//#pragma GCC push_options
//#pragma GCC optimize ("O0")

void PeriodicTimerTests::shouldBeAbleToRestartTimer()
{
    m_timer.start(100, boost::bind(&PeriodicTimerTests::Tick, this));
    m_timer.start(100, boost::bind(&PeriodicTimerTests::Tick, this));
}

//#pragma GCC pop_options

int main(int argc, char* argv[])
{
    PeriodicTimerTests foo;
    foo.shouldBeAbleToRestartTimer();

    return 0;
}
